﻿using System;
using System.Collections.Generic;
using System.Linq;
using Celeste;
using Celeste.Mod;
using Microsoft.Xna.Framework;
using Mono.Cecil.Cil;
using Monocle;
using MonoMod.Cil;

namespace TileLayoutAnchor;

using TileSubstitution = TileLayoutAnchorMapDataProcessor.TileSubstitution;

public class TileLayoutAnchorModule : EverestModule{
    public static TileLayoutAnchorModule Instance { get; private set; }

    // just to stay safe
    private const int GreatestScan = 5;
    
    public TileLayoutAnchorModule() =>
        Instance = this;

    public override void PrepareMapDataProcessors(MapDataFixup context) =>
        context.Add<TileLayoutAnchorMapDataProcessor>();

    public override void Load(){
        // effect control
        On.Monocle.ParticleSystem.Clear += PreserveParticles;
        On.Celeste.TrailManager.Clear += PreserveTrail;
        On.Celeste.DisplacementRenderer.Clear += PreserveDisplacement;
        On.Celeste.Level.LoadLevel += AdjustPreservedData;
        On.Celeste.Level.UnloadLevel += RememberLastLevel;
        
        // decal control
        On.Celeste.Decal.Added += MakeDecalStartConsistent;
        On.Celeste.Decal.MakeBanner += MakeDecalBannerConsistent;
        
        // tile control
        IL.Celeste.LevelLoader.LoadingThread += HookLevelLoad;
    }

    public override void Unload(){
        On.Monocle.ParticleSystem.Clear -= PreserveParticles;
        On.Celeste.TrailManager.Clear -= PreserveTrail;
        On.Celeste.DisplacementRenderer.Clear -= PreserveDisplacement;
        On.Celeste.Level.LoadLevel -= AdjustPreservedData;
        On.Celeste.Level.UnloadLevel -= RememberLastLevel;
        
        On.Celeste.Decal.Added -= MakeDecalStartConsistent;
        On.Celeste.Decal.MakeBanner -= MakeDecalBannerConsistent;
        
        IL.Celeste.LevelLoader.LoadingThread -= HookLevelLoad;
    }

    public static void LogDebug(string info) =>
        Logger.Log(LogLevel.Debug, "TileLayoutAnchor", info);
    
    private static void PreserveParticles(On.Monocle.ParticleSystem.orig_Clear orig, ParticleSystem self){
        if(self.Scene is Level l && l.Tracker.GetEntity<Player>() is { Dead: false } p)
            if(SubstitutionFor(p) is { ControlEffects: true })
                return;

        orig(self);
    }
    
    private static void PreserveTrail(On.Celeste.TrailManager.orig_Clear orig){
        if(Engine.Scene is Level l && l.Tracker.GetEntity<Player>() is { Dead: false } p)
            if(SubstitutionFor(p) is { ControlEffects: true })
                return;
        
        orig();
    }

    // TODO: DisplacementRenderer.Clear can get inlined, do bait&switch instead
    // (or ask popax to fix it lol)
    private static void PreserveDisplacement(On.Celeste.DisplacementRenderer.orig_Clear orig, DisplacementRenderer self){
        if(Engine.Scene is Level l && l.Tracker.GetEntity<Player>() is { Dead: false } p)
            if(SubstitutionFor(p) is { ControlEffects: true })
                return;

        orig(self);
    }
    
    private static string prevLevel = null;
    private static void AdjustPreservedData(On.Celeste.Level.orig_LoadLevel orig, Level self, Player.IntroTypes intro, bool fromLoader){
        orig(self, intro, fromLoader);
        
        if(self.Transitioning || fromLoader || prevLevel == null || self.Tracker.GetEntity<Player>() is not { Dead: false })
            return;

        if(TileLayoutAnchorMapDataProcessor.Substitutions.TryGetValue(self.Session.Area.GetSID(), out var subs)
           && (TileSubstitution?)subs.Find(x => x.Room == prevLevel) is {} prevAnchor
           && (TileSubstitution?)subs.Find(x => x.Room == self.Session.Level) is {} curAnchor
           && self.Session.MapData.Levels.Find(x => x.Name == prevLevel) is {} prevData
           && self.Session.MapData.Levels.Find(x => x.Name == self.Session.Level) is {} curData){
            Vector2 offset = (curData.Position - prevData.Position) + (curAnchor.AnchorOffset - prevAnchor.AnchorOffset);
            foreach(ParticleSystem pts in new List<ParticleSystem>{ self.Particles, self.ParticlesFG, self.ParticlesBG })
                for(int i = 0; i < pts.particles.Length; i++)
                    pts.particles[i].Position += offset;
            if(self.Tracker.GetEntity<TrailManager>() is { snapshots: {} snapshots })
                foreach(var t in snapshots)
                    if(t != null)
                        t.Position += offset;
            foreach(DisplacementRenderer.Burst b in self.Displacement.points)
                b.Position += offset;
            foreach(Entity e in self.Tracker.GetEntities<SlashFx>()){
                e.RemoveTag(Tags.Global);
                e.Position += offset;
            }
        }
    }
    
    private static void RememberLastLevel(On.Celeste.Level.orig_UnloadLevel orig, Level self){
        if(self.Tracker.GetEntity<Player>() is { Dead: false } p)
            if(SubstitutionFor(p) is { ControlEffects: true })
                foreach(Entity e in self.Tracker.GetEntities<SlashFx>())
                    e.AddTag(Tags.Global);
        
        orig(self);
        prevLevel = self.Session.Level;
    }
    
    private static void MakeDecalStartConsistent(On.Celeste.Decal.orig_Added orig, Decal self, Scene scene){
        orig(self, scene);
        if(SubstitutionFor(self) is { ControlDecals: true })
            if(self.textures.Count > 1){
                self.frame = scene.TimeActive * self.AnimationSpeed;
                self.frame %= self.textures.Count;
            }
    }
    
    private static void MakeDecalBannerConsistent(On.Celeste.Decal.orig_MakeBanner orig, Decal self, float speed, float amplitude, int sliceSize, float sliceSinIncrement, bool easeDown, float offset, bool onlyIfWindy){
        orig(self, speed, amplitude, sliceSize, sliceSinIncrement, easeDown, offset, onlyIfWindy);
        if(SubstitutionFor(self) is { ControlDecals: true })
            if(self.Get<Decal.Banner>() is Decal.Banner b)
                b.sineTimer = self.Scene.TimeActive;
    }
    
    private static void HookLevelLoad(ILContext il){
        LogDebug("Applying hook to level load!");
        
        ILCursor cursor = new(il);
        // background tiles
        cursor.GotoNext(MoveType.After, instr => instr.MatchNewobj<BackgroundTiles>());
        cursor.Emit(OpCodes.Dup); // new BackgroundTiles
        cursor.Emit(OpCodes.Ldarg_0); // LevelLoader this
        cursor.Emit(OpCodes.Ldloc_S, (byte)13); // VirtualMap<char> data1
        cursor.EmitDelegate<Action<BackgroundTiles, LevelLoader, VirtualMap<char>>>((tiles, loader, map) => AlterSolidTiles(loader, tiles.Tiles, map, GFX.BGAutotiler));
        // foreground tiles
        cursor.GotoNext(MoveType.After, instr => instr.MatchNewobj<SolidTiles>());
        cursor.Emit(OpCodes.Dup); // new SolidTiles
        cursor.Emit(OpCodes.Ldarg_0); // LevelLoader this
        cursor.Emit(OpCodes.Ldloc_S, (byte)14); // VirtualMap<char> data2
        cursor.EmitDelegate<Action<SolidTiles, LevelLoader, VirtualMap<char>>>((tiles, loader, map) => AlterSolidTiles(loader, tiles.Tiles, map, GFX.FGAutotiler));
    }
    
    private static void AlterSolidTiles(LevelLoader loader, TileGrid grid, VirtualMap<char> mapTiles, Autotiler autotiler){
        LogDebug($"Altering level load for {loader.Level.Session.Area.GetSID()}!");
        
        // swap out rooms as we deem necessary
        Level level = loader.Level;
        Session session = level.Session;
        MapData mapData = session.MapData;
        LogDebug(string.Join("; ", TileLayoutAnchorMapDataProcessor.Substitutions.Select(pair => pair.Key.ToString() + " -> " + string.Join(", ", pair.Value.Select(x => x.ToString())))));
        if(TileLayoutAnchorMapDataProcessor.Substitutions.TryGetValue(session.Area.GetSID(), out var subs)){
            LogDebug($"Found {subs.Count} substitutions!");
            // `data` refers to the entire map's tiles.
            // for each room, we want to cut out the area around a room, stable autotile that, then substitute in the centre
            foreach(var sub in subs.Where(x => x.ControlTiles)){
                if(mapData.Levels.Find(x => x.Name == sub.Room) is {} levelData){
                    // copy the relevant tiles
                    int roomX = levelData.Bounds.X / 8,
                        roomY = levelData.Bounds.Y / 8,
                        w = levelData.Bounds.Width / 8 + GreatestScan * 2,
                        h = levelData.Bounds.Height / 8 + GreatestScan * 2,
                        vXO = -(int)(sub.AnchorOffset.X / 8),
                        vYO = -(int)(sub.AnchorOffset.Y / 8);
                    VirtualMap<char> context = new VirtualMap<char>(w, h, '0');
                    for(int x = 0; x < w; x++)
                        for(int y = 0; y < h; y++)
                            context[x, y] = mapTiles[x + roomX - GreatestScan - mapData.TileBounds.X, y + roomY - GreatestScan - mapData.TileBounds.Y];
                    LogDebug($"it's like {context.Columns}x{context.Rows} in {levelData.Name} (here), shifted by ({vXO},{vYO})");
                    // autotile that
                    TileGrid roomGrid = autotiler.GenerateMapStable(context, new(), vXO, vYO).TileGrid;
                    // ignore the (invalid) context
                    roomGrid.Extend(-GreatestScan, -GreatestScan, -GreatestScan, -GreatestScan);
                    // substitute that back in
                    LogDebug($"oh you know, {mapData.TileBounds}");
                    grid.Substitute(roomGrid, roomX - mapData.TileBounds.X, roomY - mapData.TileBounds.Y);
                }
            }
        }
    }

    public static TileSubstitution? SubstitutionFor(Entity e) => SubstitutionFor(e.SceneAs<Level>().Session);

    public static TileSubstitution? SubstitutionFor(Session s){
        if(TileLayoutAnchorMapDataProcessor.Substitutions.TryGetValue(s.Area.GetSID(), out var subs))
            foreach(TileSubstitution ts in subs)
                if(ts.Room == s.Level)
                    return ts;

        return null;
    }
}